// UTILS
/**
 * @returns {HTMLElement[]} the infobox slots and inserts the ibox to the DOM
 */
ViewFrame.initBox = function (name, type, namesFrom) {
	const ibox = document.createElement('details');
	ibox.classList.add('box', type);
	const summ = document.createElement('summary');
	summ.innerText = name;
	const table = document.createElement('table');
	table.classList.add('content');
	ibox.append(summ, table);

	const iboxSlots = new Array();
	namesFrom.forEach((item, i) => {
			const slot = document.createElement('tr');
			const head = document.createElement('th');
			head.innerText = item;
			const data = document.createElement('td');
			const row = document.createElement('input');
			row.type = 'text';
			row.readOnly = true;
			iboxSlots.push(row);
			data.append(row);
			slot.append(head, data);
			table.append(slot);
		});
	document.getElementsByTagName('body')[0].append(ibox);
	return iboxSlots;
};

ViewFrame.prototype.updateEdges = function () {
	if (this.width != window.innerWidth || this.height != window.innerHeight) {
		this.windowWidth = window.innerWidth;
		this.map.canvas.width = this.windowWidth;
		this.windowHeight = window.innerHeight;
		this.map.canvas.height = this.windowHeight;
		this.map.imageSmoothingEnabled = false;
	}
	this.width = this.windowWidth / this.zoom; // in chunks
	this.height = this.windowHeight / this.zoom;

	this.east = this.x + this.width; // this is fractional
	this.south = this.y + this.height;
};

// LOCATION CHECKS
/**
 * whether or not the chunk is on screen
 * if ANY part of the chunk is visible
 */
ViewFrame.prototype.isChunkWiseOnScreen = function (coord) {
	return coord[0] >= this.x - 1  &&
		coord[0] <= this.east  &&
		coord[1] >= this.y - 1  &&
		coord[1] <= this.south ;
};

/**
 * whether or not the blockwise pos is on screen.
 */
ViewFrame.prototype.isBlockWiseOnScreen = function (coord) {
	let x = coord[0] / this.chunkSize;
	let y = coord[1] / this.chunkSize;
	return this.isChunkWiseOnScreen([x, y]);
};

// COORD CONVERTERS
/**
 * converts block x and y to screen relative coords.
 * it can return a location that IS NOT on the screen!
 */
ViewFrame.prototype.blockWiseToScreen = function (coord) {
	let x = coord[0] / this.chunkSize;
	let y = coord[1] / this.chunkSize;
	return [
		(x - this.x) * this.zoom,
		(y - this.y) * this.zoom
	];
};

/**
 * converts screen x and y to block absolute coords.
 * to do
 */
ViewFrame.prototype.screenWiseToBlocks = function (coord) {
	const [x, y] = this.screenWiseToChunks(coord);
	return [x * this.chunkSize, y * this.chunkSize];
};

/**
 * coords of chunk relative to screen
 */
ViewFrame.prototype.chunkWiseToScreen = function (coord) {
	return [
		(coord[0] - this.x) * this.zoom,
		(coord[1] - this.y) * this.zoom
	];
};

/**
 * screen coord to chunk coord
 */
ViewFrame.prototype.screenWiseToChunks = function (coord) {
	return [
		this.x + coord[0] / this.zoom,
		this.y + coord[1] / this.zoom
	];
};

// MUTATORS
/**
 * moves the center of the map.
 */
ViewFrame.prototype.moveScreenWiseCenter = function (click) {
	// to pan when we click on the map!
	this.x += (click[0] - this.windowWidth / 2) / this.zoom;
	this.y += (click[1] - this.windowHeight / 2) / this.zoom;
};

/**
 * gets/sets the map center to the given chunk coords.
 */
Object.defineProperty(ViewFrame.prototype, 'chunkWiseCenter', {
	get() {
		return [
			this.x + this.width / 2,
			this.y + this.height / 2
		]
	},
	set(coord) {
		this.x = coord[0] - this.width / 2;
		this.y = coord[1] - this.height / 2;
	}
});

/**
 * gets/sets the center of the map to be the given absolute block coords.
 */
Object.defineProperty(ViewFrame.prototype, 'blockWiseCenter', {
	get() {
		return [
			(this.x - this.width / 2) / this.chunkSize,
			(this.y - this.height / 2) / this.chunkSize
		]
	},
	set(coord) {
		this.x = coord[0] / this.chunkSize + this.width / 2;
		this.y = coord[1] / this.chunkSize + this.height / 2;
	}
});

Object.defineProperty(ViewFrame.prototype, 'zoomIndex', {
	get() {
		return ViewFrame.ZOOM_LEVELS.indexOf(this.zoom);
	},
	set(zoomIndex) {
		if (zoomIndex < ViewFrame.ZOOM_LEVELS.length - 1 && zoomIndex > 0) zoomIndex = zoomIndex;
		else zoomIndex = Math.max(0, Math.min(zoomIndex, ViewFrame.ZOOM_LEVELS.length));
		const savedCenter = this.chunkWiseCenter;
		this.zoom = ViewFrame.ZOOM_LEVELS[zoomIndex];
		this.updateEdges();
		this.chunkWiseCenter = savedCenter;
	}
});

Object.defineProperty(ViewFrame, 'ZOOM_LEVELS', {
	value: [128, 64, 32, 16, 8, 4, 2, 1]
});

function decode(img, cb) {
	img.decode()
		.then(() => {
			cb(img);
		})
		.catch(() => {}); // so images arent added on error
}

function guessNear(coord, source, dist) {
	const output = [];
	for (var x = coord[0] - dist; x < coord[0] + dist; x++) {
		for (var y = coord[1] - dist; y < coord[1] + dist; y++) {
			const name = x + '_' + y;
			if (source.has(name)) output.push([name, Math.hypot(x - coord[0], y - coord[1])]);
		}
	}
	output.sort((p1, p2) => p1[1] - p2[1] );
	return output;
}